"""Synthetic datasets for testing and examples."""

from ._composite_effect_example import load_composite_effect_example
from ._squared_exogenous import load_synthetic_squared_exogenous

__all__ = ["load_composite_effect_example", "load_synthetic_squared_exogenous"]
